/* -*- c-basic-offset: 4 -*- */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>

#include "server.h"
#include "net.h"
#include "game.h"
#include "asciiart.h"

typedef struct {
    int players[2];
    game_t game;
} server_game_t;

server_game_t *server_games[MAX_GAMES];
int server_games_sz;

#define MAX_CLIENT_NAME_LEN 10
typedef struct {
    int connected;
    char name[MAX_CLIENT_NAME_LEN];
    int game;
    int requested_game, requested_game_with;
} server_client_t;

server_client_t server_clients[MAX_CLIENTS];

typedef int cmdfunc_t(int client, char **args);
typedef struct {
    char *name;
    char *desc;
    int argc;
    cmdfunc_t *func;
} server_cmd_t;

int server_do_command(int clientid, char *cmd);

int
server_new_game(int silver, int red)
{
    if (server_games_sz == MAX_GAMES)
	return -1;
    int game_id = server_games_sz++;
    server_game_t *g = malloc(sizeof(server_game_t));

    g->players[SILVER] = silver;
    g->players[RED] = red;
    server_clients[silver].game = game_id;
    server_clients[red].game = game_id;
    server_clients[silver].requested_game = 0;
    server_clients[red].requested_game = 0;

    init_game(&g->game);
    setup_board(&g->game, classic_setup);
    server_games[game_id] = g;
    return game_id;
}

void
server_delete_game(int id)
{
    int i;
    assert(id >= 0 && id < MAX_GAMES && server_games[id] != NULL);
    free(server_games[id]);
    server_games[id] = NULL;
    for (i = id+1; i < server_games_sz; i++)
	server_games[i-1] = server_games[i];
    server_games_sz--;
}

int
server_get_game_request(int client)
{
    int i;
    server_client_t *c = server_clients;
    int match_any = -1;
    for (i = 0; i < MAX_CLIENTS; i++) {
	if (c[i].connected &&
	    c[i].requested_game) {
	    if (c[i].requested_game_with == client)
		return i;
	    else if (c[i].requested_game_with == -1)
		match_any = i;
	}
    }
    return match_any;
}

int
server_request_game(int client, int opponent)
{
    int match = 0;
    if (server_clients[client].requested_game)
	return -2;
    if (opponent == -1) {
	opponent = server_get_game_request(client);
	if (opponent != -1) match = 1;
    } else {
	match =
	    server_clients[opponent].requested_game &&
	    (server_clients[opponent].requested_game_with == client ||
	     server_clients[opponent].requested_game_with == -1);
    }
    if (match) {
	server_clients[opponent].requested_game = 0;
	return server_new_game(client, opponent);
    } else {
	server_clients[client].requested_game = 1;
	server_clients[client].requested_game_with = opponent;
	return -1;
    }
}

int
server_client_by_name(char *name)
{
    int i;
    for (i = 0; i < MAX_CLIENTS; i++) {
	if (server_clients[i].connected &&
	    strcasecmp(server_clients[i].name, name)==0)
	    return i;
    }
    return -1;
}


void
terminal_game(void)
{
    game_t game;
    move_t move;
    char aabuf[ASCIIART_SMALL_BUFSZ];
    char linebuf[10];

    init_game(&game);
    setup_board(&game, classic_setup);
    asciiart_small_draw_board(aabuf, &game);
    printf("%s", aabuf);

    for (;;) {
	printf("move %d, %s\n", game.move, sideinfo[game_turn(&game)].name);
	fgets(linebuf, 10, stdin);
	if (parse_move(linebuf, &move) != HUGE_SUCCESS) {
	    printf("not a valid move description: '%s'\n", linebuf);
	    continue;
	}
	if (perform_move(&game, game_turn(&game), move) != HUGE_SUCCESS) {
	    printf("invalid move\n");
	    continue;
	}
	
	asciiart_small_draw_board(aabuf, &game);
	printf("%s", aabuf);

	if (game.last_move.type == M_ROTATE) {
	    printf("move: %c%c rotated %s\n",
		   game.last_move.y+'A',
		   game.last_move.x+'0',
		   game.last_move.dir==R_CW ? "R" : "L");
	} else {
	    int nx, ny;
	    move_dest(game.last_move, &nx, &ny);
	    printf("move: %c%c moved to %c%c\n",
		   game.last_move.y+'A',
		   game.last_move.x+'0',
		   ny+'A', nx+'0');
	}

	if (game.last_hit.piece != NONE) {
	    printf("laser hits %s %s at %c%c\n",
		   sideinfo[game.last_hit.side].name,
		   pieceinfo[game.last_hit.piece].name,
		   game.laser_pos[1]+'A', game.laser_pos[0]+'0');
	} else {
	    printf("laser hits wall at %c%c\n",
		   game.laser_pos[1]+'A', game.laser_pos[0]+'0');
	}

	switch (game.winner) {
	case SILVER:
	    printf("SILVER WINS!\n");
	    return;
	case RED:
	    printf("RED WINS!\n");
	    return;
	}

    }
}


void
server_report_game_state(int gameid)
{
    server_game_t *g = server_games[gameid];
    int i;
    for (i = 0; i < 2; i++) {
	int c = g->players[i];

	if (g->game.move == 0) {
	    net_client_printf(c, "started game between %s (silver) and %s (red)\n",
			      server_clients[g->players[SILVER]].name,
			      server_clients[g->players[RED]].name);
	} else {
	    if (g->game.last_move.type == M_ROTATE) {
		net_client_printf(c, "move: %c%c rotated %s\n",
				  g->game.last_move.y+'A',
				  g->game.last_move.x+'0',
				  g->game.last_move.dir==R_CW ? "R" : "L");
	    } else {
		int nx, ny;
		move_dest(g->game.last_move, &nx, &ny);
		net_client_printf(c, "move: %c%c moved to %c%c\n",
				  g->game.last_move.y+'A',
				  g->game.last_move.x+'0',
				  ny+'A', nx+'0');
	    }

	    if (g->game.last_hit.piece != NONE) {
		net_client_printf(c, "laser hits %s %s at %c%c\n",
				  sideinfo[g->game.last_hit.side].name,
				  pieceinfo[g->game.last_hit.piece].name,
				  g->game.laser_pos[1]+'A', g->game.laser_pos[0]+'0');
	    } else {
		net_client_printf(c, "laser hits wall at %c%c\n",
				  g->game.laser_pos[1]+'A', g->game.laser_pos[0]+'0');
	    }
	}

	char aabuf[ASCIIART_SMALL_BUFSZ];
	asciiart_small_draw_board(aabuf, &g->game);
	net_client_printf(c, "%s", aabuf);
	
	switch (g->game.winner) {
	case SILVER:
	    net_client_printf(c, "SILVER WINS!\n");
	    break;
	case RED:
	    net_client_printf(c, "RED WINS!\n");
	    break;
	default:
	    if (g->players[game_turn(&g->game)]==c)
		net_client_printf(c, "your turn\n");
	}
    }
}

void
server_do_move(int clientid, move_t move)
{
    int gameid = server_clients[clientid].game;
    server_game_t *g = server_games[gameid];
    int side;

    side = (g->players[1]==clientid);
    
    switch (perform_move(&g->game, side, move)) {
    case HUGE_SUCCESS:
	break;
    case E_NOT_YOUR_TURN:
	net_client_printf(clientid, "not your turn\n");
	return;
    default:
	net_client_printf(clientid, "invalid move\n");
	return;
    }

    server_report_game_state(gameid);

    switch (g->game.winner) {
    case SILVER:
	server_delete_game(gameid);
	break;
    case RED:
	server_delete_game(server_clients[clientid].game);
	break;
    }
}

void
parse_line(int clientid, char *line)
{
    server_do_command(clientid, line);
}

int player[2]={0, 0};

void
server_client_connect(int clientid) {
    server_client_t *c = &server_clients[clientid];
    c->connected = 1;
    sprintf(c->name, "client%03d", clientid);
    c->game = -1;
    c->requested_game = 0;

    net_client_printf(clientid, "Welcome to the KHET server\n");
    net_client_printf(clientid, "Say 'play' to play a game\n");
    net_client_printf(clientid, "(you may have to wait for another player)\n");

    /*
    assert(clientid<2);
    player[clientid]=1;
    if (player[0] && player[1]) {
	char aabuf[ASCIIART_SMALL_BUFSZ];
	init_game(&game);
	setup_board(&game, classic_setup);
	asciiart_small_draw_board(aabuf, &game);
	net_all_printf("%s", aabuf);
    }
    */
}


void
server_init(void)
{
    int i;
    server_games_sz = 0;
    for (i = 0; i < MAX_CLIENTS; i++)
	server_clients[i].connected = 0;
}



int
server_cmd_play(int client, char **args)
{
    int g = server_request_game(client, -1);
    if (g == -1) {
	net_client_printf(client, "requested game with anyone\n");
    } else if (g == -2) {
	net_client_printf(client, "can not request another game\n");
    } else {
	server_report_game_state(g);
    }
    return 1;
}

int
server_cmd_playwith(int client, char **args)
{
    int opponent = server_client_by_name(args[0]);
    if (opponent == -1) {
	net_client_printf(client, "there is no player named %s\n", args[0]);
	return 1;
    }

    int g = server_request_game(client, opponent);
    if (g == -1) {
	net_client_printf(client, "requested game with %s\n", args[0]);
	net_client_printf(opponent, "%s wants to play a game with you\n",
			  server_clients[client].name);
    } else if (g == -2) {
	net_client_printf(client, "can not request another game\n");
    } else {
	server_report_game_state(g);
    }
    return 1;
}

int
server_cmd_nick(int client, char **args)
{
    strncpy(server_clients[client].name, args[0], MAX_CLIENT_NAME_LEN);
    server_clients[client].name[MAX_CLIENT_NAME_LEN-1] = '\0';
    net_client_printf(client, "nick changed to %s\n",
		      server_clients[client].name);
    return 1;
}

int
server_cmd_ls(int client, char **args)
{
    char *thing = args[0];
    int i;
    if (strcasecmp(thing, "players")==0) {
	for (i = 0; i < MAX_CLIENTS; i++) {
	    if (server_clients[i].connected) {
		net_client_printf(client, "%s\n", server_clients[i].name);
	    }
	}
    } else if (strcasecmp(thing, "games")==0) {
	for (i = 0; i < server_games_sz; i++) {
	    server_game_t *g = server_games[i];
	    net_client_printf(client, "game %d: %s -- %s (%d moves)\n", i,
			      server_clients[g->players[0]].name,
			      server_clients[g->players[1]].name,
			      g->game.move);
	}
	net_client_printf(client, "%d games\n", server_games_sz);
    }
    return 1;
}

int
server_cmd_whoami(int client, char **args)
{
    net_client_printf(client, "%s\n", server_clients[client].name);
    return 1;
}

server_cmd_t server_commands[] = {
    { .name = "play",
      .desc = "play a game with anyone",
      .argc = 0,
      .func = server_cmd_play },
    { .name = "playwith",
      .desc = "play a game with a named opponent",
      .argc = 1,
      .func = server_cmd_playwith },
    { .name = "nick",
      .desc = "set your nickname",
      .argc = 1,
      .func = server_cmd_nick },
    { .name = "ls",
      .desc = "list stuff",
      .argc = 1,
      .func = server_cmd_ls },
    { .name = "whoami",
      .desc = "find out what your name is",
      .argc = 0,
      .func = server_cmd_whoami } };
int server_commands_sz = 5;

int
server_do_command(int clientid, char *cmd)
{
    char *cmd_copy = strdup(cmd);
    server_cmd_t *cmd_type = NULL;
    char *args[10];
    int argc;
    char *c;
    move_t move;
    int i;

    /*
    for (i = 0; i < strlen(cmd_copy); i++) {
	cmd_copy[i] = toupper(cmd_copy[i]);
    }
    */

    if (server_clients[clientid].game != -1 &&
	parse_move(cmd_copy, &move) == HUGE_SUCCESS) {
	server_do_move(clientid, move);
	return 1;
    }

    for (argc = 0, c = cmd_copy; ; argc++) {
	while (*c != ' ' && *c != '\n' && *c != '\0')
	    c++;
	if (*c == ' ' || *c == '\n') *(c++) = '\0';
	while (*c == ' ' || *c == '\n')
	    c++;
	if (*c == '\n' || *c == '\0')
	    break;
	args[argc] = c;
    }

    for (i = 0; i < server_commands_sz; i++) {
	if (strcasecmp(cmd_copy, server_commands[i].name)==0)
	    cmd_type = &server_commands[i];
    }
    if (cmd_type == NULL) {
	net_client_printf(clientid, "unknown command %s\n", cmd_copy);
	return 0;
    }
    if (argc != cmd_type->argc) {
	net_client_printf(clientid,
			  "wrong number of arguments for %s (wanted %d, got %d)\n",
			  cmd_type->name, cmd_type->argc, argc);
	return 0;
    }

    return cmd_type->func(clientid, args);
}

int
main(int argc, char **argv) {
    if (argc > 1 && strcmp(argv[1], "local")==0) {
	terminal_game();
	return 0;
    } else {
	net_server(15239);
	for (;;)
	    net_poll();
    }
    return 0;
}
